# =============================================================================
# Minet CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
from typing import Optional, Iterable, List, Iterator

import os
import sys
import stat
import yaml
import platform
import casanova
from casanova.namedrecord import is_tabular_record_class
from dataclasses import dataclass
from glob import iglob
from copy import copy
from os.path import join, expanduser, isfile, relpath
from collections.abc import Mapping
from functools import wraps
from logging import Handler
from contextlib import nullcontext
from ebbe import format_seconds, get
from types import SimpleNamespace

from minet.crawl import CrawlerState
from minet.encodings import is_supported_encoding
from minet.cli.console import console
from minet.cli.loading_bar import LoadingBar, StatsItem
from minet.cli.exceptions import FatalError
from minet.utils import fuzzy_int, message_flatmap


def colored(string, color):
    return "[{color}]{string}[/{color}]".format(string=string, color=color)


def redirect_to_devnull():
    # Taken from: https://docs.python.org/3/library/signal.html
    devnull = os.open(os.devnull, os.O_WRONLY)
    os.dup2(devnull, sys.stdout.fileno())


def with_cli_exceptions(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            fn(*args, **kwargs)

        except KeyboardInterrupt:
            redirect_to_devnull()
            sys.exit(1)

        except BrokenPipeError:
            console.print(
                "[warning]minet process was stopped because piped command exited!"
            )
            redirect_to_devnull()
            sys.exit(1)

        except Exception:
            console.print("[error]minet process was stopped because an error occurred!")
            raise

    return wrapper


def get_stdin_status():
    mode = os.fstat(sys.stdin.fileno()).st_mode

    if stat.S_ISFIFO(mode):
        return "piped"
    elif stat.S_ISREG(mode):
        return "redirected"

    return "terminal"


def is_stdin_empty():
    return not sys.stdin.buffer.peek(1)


def was_piped_something():
    return get_stdin_status() != "terminal" and not is_stdin_empty()


def print_err(*messages):
    print(message_flatmap(*messages), file=sys.stderr)


def die(*messages):
    print_err(*messages)

    sys.exit(1)


def safe_index(l, e):
    try:
        return l.index(e)
    except ValueError:
        return None


class CLIRetryerHandler(Handler):
    def emit(self, record):
        if record.source == "request_retryer":
            exc = record.exception
            pretty_time = format_seconds(record.sleep_time)

            exc_name = "%s.%s" % (exc.__class__.__module__, exc.__class__.__name__)
            exc_msg = str(exc)

            msg = [
                "%s now wait for %s because of following exception:"
                % ("A thread will" if record.from_thread else "Will", pretty_time),
                exc_name,
            ]

            if exc_msg:
                msg.append("Exception message: %s" % exc_msg)

            if record.epilog:
                msg.append("[dim]%s[/dim]" % record.epilog)

            msg.append("")

        else:
            msg = [record.msg, ""]

        console.log_with_time("\n".join(msg), style="warning")


def acquire_cross_platform_stdout():
    # As per #254: stdout need to be wrapped so that windows get a correct csv
    # stream output
    if "windows" in platform.system().lower():
        return open(
            sys.__stdout__.fileno(),
            mode=sys.__stdout__.mode,
            buffering=1,
            encoding=sys.__stdout__.encoding,
            errors=sys.__stdout__.errors,
            newline="",
            closefd=False,
        )

    return sys.stdout


@dataclass
class FetchReportLikeItem:
    index: int
    row: List[str]
    path: Optional[str] = None
    encoding: Optional[str] = None
    text: Optional[str] = None
    error: Optional[str] = None
    url: Optional[str] = None


def create_fetch_like_report_iterator(
    cli_args: SimpleNamespace, reader: casanova.Reader
) -> Iterator[FetchReportLikeItem]:
    headers = reader.headers
    input_dir = cli_args.input_dir or ""

    # TODO: deal with no_headers
    assert headers is not None

    url_column = getattr(cli_args, "url_column", None)

    filename_pos = headers.get(cli_args.column) if cli_args.column is not None else None
    error_pos = headers.get(cli_args.error_column)
    status_pos = headers.get(cli_args.status_column)
    encoding_pos = headers.get(cli_args.encoding_column)
    mimetype_pos = headers.get(cli_args.mimetype_column)
    body_pos = headers.get(cli_args.body_column)
    url_pos = headers.get(url_column) if url_column is not None else None

    for i, row in reader.enumerate():
        item = FetchReportLikeItem(index=i, row=row)

        if url_pos is not None:
            url = get(row, url_pos, "").strip()

            if url:
                item.url = url

        if error_pos is not None:
            error = get(row, error_pos, "").strip()

            if error:
                item.error = "http-errored"
                yield item
                continue

        if status_pos is not None:
            status = get(row, status_pos)

            if status:
                status = fuzzy_int(status)

            if status != 200:
                item.error = "invalid-status"
                yield item
                continue

        if mimetype_pos is not None:
            mimetype = get(row, mimetype_pos, "text/html").strip()

            if "/htm" not in mimetype:
                item.error = "invalid-mimetype"
                yield item
                continue

        if filename_pos is not None:
            filename = get(row, filename_pos, "").strip()

            if filename:
                item.path = join(input_dir, filename)

        if body_pos is not None:
            body = get(row, body_pos)

            if body is not None:
                item.text = body

        if item.path is None and item.text is None:
            item.error = "no-filename-nor-body"
            yield item
            continue

        # NOTE: can be None, which means we will guess
        item.encoding = cli_args.encoding

        if encoding_pos is not None:
            encoding = get(row, encoding_pos, "").strip()

            if encoding:
                if not is_supported_encoding(encoding):
                    item.error = "encoding-not-supported"
                    yield item
                    continue

                item.encoding = encoding

        if cli_args.glob:
            for p in iglob(item.path, recursive=True):
                new_item = copy(item)
                new_item.path = relpath(p, start=input_dir)
                yield new_item

            continue

        yield item


def get_rcfile(rcfile_path=None):
    home = expanduser("~")

    files = [
        rcfile_path,
        ".minetrc",
        ".minetrc.yml",
        ".minetrc.yaml",
        ".minetrc.json",
        join(home, ".minetrc"),
        join(home, ".minetrc.yml"),
        join(home, ".minetrc.yaml"),
        join(home, ".minetrc.json"),
    ]

    for p in files:
        if p is None or not isfile(p):
            continue

        with open(p, encoding="utf-8") as f:
            return yaml.safe_load(f)

    return None


def with_fatal_errors(mapping_or_hook):
    if not isinstance(mapping_or_hook, Mapping) and not callable(mapping_or_hook):
        raise TypeError("Expecting mapping or callable")

    def decorate(action):
        @wraps(action)
        def wrapper(*args, **kwargs):
            try:
                return action(*args, **kwargs)
            except Exception as e:
                msg = None

                if isinstance(mapping_or_hook, Mapping):
                    msg = mapping_or_hook.get(type(e))
                elif callable(mapping_or_hook):
                    msg = mapping_or_hook(args[0], e)

                if msg is not None:
                    raise FatalError(msg)

                raise e

        return wrapper

    return decorate


def with_loading_bar(stats: Optional[Iterable[StatsItem]] = None, **loading_bar_kwargs):
    def decorate(action):
        @wraps(action)
        def wrapper(cli_args, *args, **kwargs):
            total = getattr(cli_args, "total", None)

            with LoadingBar(
                total=total,
                stats=stats,
                refresh_per_second=getattr(cli_args, "refresh_per_second", 10),
                single_line=getattr(cli_args, "single_line", False),
                **loading_bar_kwargs
            ) as loading_bar:
                additional_kwargs = {
                    "loading_bar": loading_bar,
                }

                return action(cli_args, *args, **additional_kwargs, **kwargs)

        return wrapper

    return decorate


def with_enricher_and_loading_bar(
    headers,
    enricher_type=None,
    get_input=None,
    index_column: Optional[str] = None,
    #
    title=None,
    unit: Optional[str] = None,
    sub_unit: Optional[str] = None,
    stats: Optional[Iterable[StatsItem]] = None,
    stats_sort_key=None,
    nested: bool = False,
    show_label: bool = False,
    total_from_enricher: bool = True,
):
    def decorate(action):
        @wraps(action)
        def wrapper(cli_args, *args, **kwargs):
            enricher_context = nullcontext()

            # Do we need to display a transient resume progress?
            if (
                hasattr(cli_args, "resume")
                and cli_args.resume
                and isinstance(
                    cli_args.output,
                    (casanova.RowCountResumer, casanova.ThreadSafeResumer),
                )
                and cli_args.output.can_resume()
            ):
                resume_loading_bar = LoadingBar(
                    title="Reading output to resume",
                    unit="lines",
                    transient=True,
                    refresh_per_second=getattr(cli_args, "refresh_per_second", 10),
                    single_line=getattr(cli_args, "single_line", False),
                )
                enricher_context = resume_loading_bar

                def listener(event, _):
                    if event == "output.row.read":
                        resume_loading_bar.advance()

                cli_args.output.set_listener(listener)

            enricher_fn = casanova.enricher

            if enricher_type == "threadsafe":
                enricher_fn = casanova.threadsafe_enricher

            elif enricher_type == "batch":
                enricher_fn = casanova.batch_enricher

            elif enricher_type is not None:
                raise TypeError("wrong enricher type")

            enricher_kwargs = {}

            if index_column is not None:
                enricher_kwargs["index_column"] = index_column

            multiplex = None

            if getattr(cli_args, "explode", None) is not None:
                multiplex = casanova.Multiplexer(cli_args.column, cli_args.explode)

            with enricher_context:
                enricher = enricher_fn(
                    cli_args.input if not callable(get_input) else get_input(cli_args),
                    cli_args.output,
                    add=headers(cli_args)
                    if callable(headers) and not is_tabular_record_class(headers)
                    else headers,
                    select=cli_args.select,
                    total=getattr(cli_args, "total", None),
                    multiplex=multiplex,
                    **enricher_kwargs
                )

            completed = 0

            if isinstance(cli_args.output, casanova.Resumer):
                try:
                    completed = cli_args.output.already_done_count()
                except NotImplementedError:
                    pass

            with LoadingBar(
                title=title(cli_args) if callable(title) else title,
                total=enricher.total if total_from_enricher else None,
                unit=unit(cli_args) if callable(unit) else unit,
                sub_unit=sub_unit(cli_args) if callable(sub_unit) else sub_unit,
                nested=nested,
                stats=stats,
                stats_sort_key=stats_sort_key,
                show_label=show_label,
                completed=completed,
                refresh_per_second=getattr(cli_args, "refresh_per_second", 10),
                single_line=getattr(cli_args, "single_line", False),
            ) as loading_bar:
                additional_kwargs = {
                    "enricher": enricher,
                    "loading_bar": loading_bar,
                }

                return action(cli_args, *args, **additional_kwargs, **kwargs)

        return wrapper

    return decorate


def with_ctrl_c_warning(fn):
    def wrapper(cli_args, loading_bar, **kwargs):
        try:
            fn(cli_args, loading_bar=loading_bar, **kwargs)
        except KeyboardInterrupt:
            loading_bar.cursor_up()
            loading_bar.stop()
            console.print("Performing clean shutdown by cancelling ongoing calls...")
            console.print("This may take some seconds if you are hitting slow servers.")
            console.print("Ctrl-C again if you want to force exit.")
            raise

    return wrapper


def track_crawler_state_with_loading_bar(
    loading_bar: LoadingBar, crawler_state: CrawlerState
) -> None:
    def on_state_update(state: CrawlerState):
        loading_bar.set_total(state.total)
        loading_bar.set_stat("queued", state.jobs_queued)
        loading_bar.set_stat("doing", state.jobs_doing)
        loading_bar.set_stat("done", state.jobs_done)

    crawler_state.set_listener(on_state_update)



# =============================================================================
# Minet Youtube Comments CLI Action
# =============================================================================
#
# Action retrieving the comments of YouTube videos using the API.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.youtube import YouTubeAPIClient
from minet.youtube.constants import YOUTUBE_COMMENT_CSV_HEADERS
from minet.youtube.exceptions import (
    YouTubeDisabledCommentsError,
    YouTubeVideoNotFoundError,
    YouTubeExclusiveMemberError,
    YouTubeUnknown403Error,
)


@with_enricher_and_loading_bar(
    headers=YOUTUBE_COMMENT_CSV_HEADERS,
    title="Collecting video comments",
    unit="videos",
    sub_unit="comments",
    nested=True,
    stats=[
        {"name": "disabled", "style": "warning"},
        {"name": "not-found", "style": "error"},
        {"name": "exclusive-member", "style": "warning"},
        {"name": "403", "style": "error"},
    ],
)
def action(cli_args, enricher, loading_bar):
    client = YouTubeAPIClient(cli_args.key)

    for row, video in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(video):
            generator = client.comments(video)

            try:
                for comment in generator:
                    enricher.writerow(row, comment.as_csv_row())
                    loading_bar.nested_advance()

            except YouTubeDisabledCommentsError:
                loading_bar.inc_stat("disabled")

            except YouTubeVideoNotFoundError:
                loading_bar.inc_stat("not-found")

            except YouTubeExclusiveMemberError:
                loading_bar.inc_stat("exclusive-member")

            except YouTubeUnknown403Error:
                loading_bar.inc_stat("403")



# =============================================================================
# Minet Hyphe Destroy CLI Action
# =============================================================================
#
# Logic of the `hyphe destroy` action.
#
from minet.hyphe import HypheAPIClient

from minet.cli.hyphe.utils import with_hyphe_fatal_errors


@with_hyphe_fatal_errors
def action(cli_args):
    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)
    corpus.ensure_is_started()

    corpus.call("force_destroy_corpus")

    print("Corpus was successfully destroyed!")



# =============================================================================
# Minet Tiktok Formatters
# =============================================================================
#
# Various data formatters for Tiktok data.
#
from casanova import namedrecord
from ebbe import getpath
from minet.dates import timestamp_to_isoformat

from minet.tiktok.constants import (
    TIKTOK_VIDEO_CSV_HEADERS,
)

TiktokVideo = namedrecord(
    "TiktokVideo",
    TIKTOK_VIDEO_CSV_HEADERS,
    boolean=[
        "original_item",
        "author_verified",
        "author_secret",
        "music_original",
        "duet_enabled",
        "stitch_enabled",
        "share_enabled",
        "is_ad",
    ],
    plural=[
        "challenges_titles",
        "challenges_descriptions",
        "stickers_texts",
        "effect_stickers_names",
    ],
)


def format_video(item):
    challenges_titles = []
    challenges_descs = []
    stickers_texts = []
    effect_stickers_names = []

    challenges = item.get("challenges")
    if challenges:
        for challenge in challenges:
            challenges_titles.append(challenge.get("title"))
            challenges_descs.append(challenge.get("desc"))

    stickers = item.get("stickersOnItem")
    if stickers:
        for sticker in stickers:
            stickers_texts.append(" ".join(sticker.get("stickerText")))

    effects = item.get("effectStickers")
    if effects:
        for effect in effects:
            effect_stickers_names.append(effect.get("name"))

    duet_from_id = getpath(item, ["duetInfo", "duetFromId"])

    if duet_from_id == "0":
        duet_from_id = None
        duet_from_user_id = None
        duet_from_user_name = None
        duet_from_user_is_commerce = None
    else:
        for text in getpath(item, ["textExtra"]):
            if text["awemeId"] == duet_from_id:
                duet_from_user_id = text["userId"]
                duet_from_user_name = text["userUniqueId"]
                duet_from_user_is_commerce = text["isCommerce"]

    row = TiktokVideo(
        item.get("id"),
        item.get("desc"),
        item.get("createTime"),
        timestamp_to_isoformat(item.get("createTime")),
        item.get("originalItem"),
        getpath(item, ["video", "id"]),
        getpath(item, ["video", "height"]),
        getpath(item, ["video", "width"]),
        getpath(item, ["video", "duration"]),
        getpath(item, ["video", "ratio"]),
        getpath(item, ["video", "cover"]),
        getpath(item, ["video", "originCover"]),
        getpath(item, ["video", "dynamicCover"]),
        # Access denied
        # getpath(item, ["video", "downloadAddr"]),
        getpath(item, ["video", "reflowCover"]),
        getpath(item, ["video", "bitrate"]),
        getpath(item, ["video", "format"]),
        getpath(item, ["video", "videoQuality"]),
        getpath(item, ["author", "id"]),
        getpath(item, ["author", "uniqueId"]),
        getpath(item, ["author", "nickname"]),
        getpath(item, ["author", "avatarLarger"]),
        getpath(item, ["author", "signature"]),
        getpath(item, ["author", "verified"]),
        getpath(item, ["author", "secret"]),
        getpath(item, ["authorStats", "followingCount"]),
        getpath(item, ["authorStats", "followerCount"]),
        getpath(item, ["authorStats", "heartCount"]),
        getpath(item, ["authorStats", "videoCount"]),
        getpath(item, ["authorStats", "diggCount"]),
        getpath(item, ["music", "id"]),
        getpath(item, ["music", "title"]),
        getpath(item, ["music", "authorName"]),
        getpath(item, ["music", "original"]),
        getpath(item, ["music", "playUrl"]),
        getpath(item, ["music", "duration"]),
        getpath(item, ["music", "album"]),
        getpath(item, ["stats", "diggCount"]),
        getpath(item, ["stats", "shareCount"]),
        getpath(item, ["stats", "commentCount"]),
        getpath(item, ["stats", "playCount"]),
        duet_from_id,
        duet_from_user_id,
        duet_from_user_name,
        duet_from_user_is_commerce,
        challenges_titles,
        challenges_descs,
        stickers_texts,
        effect_stickers_names,
        item.get("duetEnabled"),
        item.get("stitchEnabled"),
        item.get("shareEnabled"),
        item.get("isAd"),
    )

    return row



# =============================================================================
# Minet Mediacloud Formatters
# =============================================================================
#
# Miscellaneous mediacloud-related formatter functions.
#
from casanova import namedrecord

from minet.mediacloud.utils import explode_tags
from minet.mediacloud.constants import (
    MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS,
    MEDIACLOUD_STORIES_CSV_HEADER,
    MEDIACLOUD_MEDIA_CSV_HEADER,
    MEDIACLOUD_FEED_CSV_HEADER,
)

MediacloudTopicStory = namedrecord(
    "MediacloudTopicStory",
    MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS,
    boolean=["full_text_rss"],
)

MediacloudStory = namedrecord(
    "MediacloudStory",
    MEDIACLOUD_STORIES_CSV_HEADER,
    plural=["tags", "tag_sets", "tags_ids", "tag_sets_ids"],
)

MediacloudMedia = namedrecord(
    "MediacloudMedia",
    MEDIACLOUD_MEDIA_CSV_HEADER,
    boolean=["is_healthy", "is_monitored"],
    plural=["tags", "tag_sets", "tags_ids", "tag_sets_ids"],
)

MediacloudFeed = namedrecord(
    "MediacloudFeed", MEDIACLOUD_FEED_CSV_HEADER, boolean=["active"]
)


def format_topic_story(story, next_link_id=None):
    return MediacloudTopicStory(
        story["guid"],
        story["stories_id"],
        story["title"],
        story["url"],
        story["language"],
        story["media_id"],
        story["media_name"],
        story["collect_date"],
        story["publish_date"],
        story["date_is_reliable"],
        story["facebook_share_count"],
        bool(story["full_text_rss"]),
        story["inlink_count"],
        story["outlink_count"],
        story["media_inlink_count"],
        story["post_count"],
        story["snapshots_id"],
        story["timespans_id"],
        next_link_id,
    )


def format_story(story, as_dict=False):
    tags = story.get("story_tags")

    return MediacloudStory(
        story["guid"],
        story["stories_id"],
        story["processed_stories_id"],
        story["title"],
        story["url"],
        story["language"],
        story["collect_date"],
        story["publish_date"],
        story["media_id"],
        story["media_name"],
        story["media_url"],
        *explode_tags(tags)
    )


def format_media(media, as_dict=False):
    tags = media.get("media_source_tags")

    return MediacloudMedia(
        media["media_id"],
        media["name"],
        media["url"],
        media["is_healthy"],
        media["is_monitored"],
        media.get("public_notes"),
        media["num_stories_90"],
        media["num_sentences_90"],
        media["start_date"],
        *explode_tags(tags)
    )


def format_feed(feed, as_dict=False):
    return MediacloudFeed(
        feed["name"],
        feed["url"],
        feed["feeds_id"],
        feed["type"],
        feed["media_id"],
        feed["active"],
        feed.get("last_attempted_download_time"),
        feed.get("last_new_story_time"),
        feed.get("last_successful_download_time"),
    )



from typing import Optional, Dict

import browser_cookie3
from urllib.request import Request
from http.cookiejar import CookieJar
from http.cookies import _unquote

from minet.exceptions import UnknownBrowserError, CookieGrabbingError
from minet.constants import COOKIE_BROWSERS


# NOTE: this code comes from Django
# ref: https://bugs.python.org/issue27674
# ref: https://github.com/django/django/commit/93a135d111c2569d88d65a3f4ad9e6d9ad291452
# NOTE: we cannot rely on python's SimpleCookie.load because it does not
# handle double quote and JSON values properly, which against the spec but happens
# in real life quite frequently.
def cookie_string_to_dict(cookie: str) -> Dict[str, str]:
    cookie_dict = {}

    for chunk in cookie.split(";"):
        if "=" in chunk:
            key, value = chunk.split("=", 1)
        else:
            key, value = "", chunk

        key = key.strip()
        value = value.strip()

        if key or value:
            cookie_dict[key] = _unquote(value)

    return cookie_dict


class CookieResolver(object):
    def __init__(self, jar: CookieJar):
        self.jar = jar

    def __call__(self, url: str) -> Optional[str]:
        req = Request(url)
        self.jar.add_cookie_header(req)

        return req.get_header("Cookie") or None


def get_cookie_jar_from_browser(browser: str = "firefox") -> CookieJar:
    fn = getattr(browser_cookie3, browser, None)

    if fn is None:
        raise UnknownBrowserError(browser)

    try:
        return fn()
    except Exception as reason:
        raise CookieGrabbingError(browser, reason=reason)


def get_cookie_resolver_from_browser(browser: str = "firefox") -> CookieResolver:
    return CookieResolver(get_cookie_jar_from_browser(browser))


def coerce_cookie_for_url_from_browser(target: str, url: str) -> Optional[str]:
    if target in COOKIE_BROWSERS:
        get = get_cookie_resolver_from_browser(target)

        return get(url)

    return target.strip()


def get_cookie_morsel_value(cookie: str, key: str) -> str:
    # TODO: could be optimized if required
    parsed = cookie_string_to_dict(cookie)
    return parsed[key]


def dict_to_cookie_string(d: Dict[str, str]) -> str:
    return "; ".join("%s=%s" % r for r in d.items())



# =============================================================================
# Minet Google Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class GoogleError(MinetError):
    pass


class GoogleSheetsError(GoogleError):
    pass


class GoogleSheetsInvalidTargetError(GoogleSheetsError):
    pass


class GoogleSheetsInvalidContentTypeError(GoogleSheetsError):
    pass


class GoogleSheetsMissingCookieError(GoogleSheetsError):
    pass


class GoogleSheetsNotFoundError(GoogleSheetsError):
    pass


class GoogleSheetsUnauthorizedError(GoogleSheetsError):
    pass


class GoogleSheetsMaxAttemptsExceeded(GoogleSheetsError):
    pass



# =============================================================================
# Minet BuzzSumo CLI Action
# =============================================================================
#
# Logic of the `bz` action.
#
from argparse import ArgumentTypeError
from datetime import datetime

from minet.cli.argparse import command, ConfigAction

FIVE_YEARS_IN_SEC = 5 * 365.25 * 24 * 60 * 60


class BuzzSumoDateType(object):
    def __call__(self, date):
        try:
            timestamp = int(datetime.strptime(date, "%Y-%m-%d").timestamp())
        except ValueError:
            raise ArgumentTypeError(
                "dates should have the following format : YYYY-MM-DD."
            )

        if (datetime.now().timestamp() - timestamp) > FIVE_YEARS_IN_SEC:
            raise ArgumentTypeError(
                "you cannot query BuzzSumo using dates before 5 years ago."
            )

        return timestamp


BUZZSUMO_LIMIT_SUBCOMMAND = command(
    "limit",
    "minet.cli.buzzsumo.limit",
    title="Minet Buzzsumo Limit Command",
    description="""
        Call BuzzSumo for a given request and return the remaining number
        of calls for this month contained in the request's headers.
    """,
    epilog="""
        Examples:

        . Returning the remaining number of calls for this month:
            $ minet bz limit --token YOUR_TOKEN
    """,
)

DATE_ARGUMENTS = [
    {
        "flag": "--begin-date",
        "help": "The date you wish to fetch articles from. UTC date should have the following format : YYYY-MM-DD",
        "required": True,
        "type": BuzzSumoDateType(),
    },
    {
        "flag": "--end-date",
        "help": "The date you wish to fetch articles to. UTC date should have the following format : YYYY-MM-DD",
        "required": True,
        "type": BuzzSumoDateType(),
    },
]

BUZZSUMO_DOMAIN_COMMAND = command(
    "domain",
    "minet.cli.buzzsumo.domain",
    title="Minet Buzzsumo Domain Command",
    description="""
        Gather social media information about all the articles crawled by BuzzSumo for one or a list of domain names and over a given period.

        The link to the official documentation: https://developers.buzzsumo.com/reference/articles.
    """,
    epilog="""
        Examples:

        . Returning social media information for one domain name:
            $ minet bz domain 'trump-feed.com' --begin-date 2021-01-01 --end-date 2021-06-30 --token YOUR_TOKEN > trump_feed_articles.csv

        . Returning social media information for a list of domain names in a CSV:
            $ minet bz domain domain_name domain_names.csv --select domain_name --begin-date 2019-01-01 --end-date 2020-12-31 --token YOUR_TOKEN > domain_name_articles.csv
    """,
    variadic_input={"dummy_column": "domain_name", "item_label": "domain name"},
    arguments=[
        *DATE_ARGUMENTS,
    ],
)


BUZZSUMO_DOMAIN_SUMMARY_COMMAND = command(
    "domain-summary",
    "minet.cli.buzzsumo.domain_summary",
    title="Minet Buzzsumo Domain Summary Command",
    description="""
        Gather information about the quantity of articles crawled by BuzzSumo for certain domain names and a given period.

        Inform the user about the number of calls (corresponding to the number of pages) needed to request BuzzSumo about those domain names.
    """,
    epilog="""
        Examples:

        . Returning the number of articles and pages found in BuzzSumo for one domain name:
            $ minet bz domain-summary 'nytimes.com' --begin-date 2019-01-01 --end-date 2019-03-01 --token YOUR_TOKEN

        . Returning the number of articles and pages found in BuzzSumo for a list of domain names in a CSV:
            $ minet bz domain-summary domain_name domain_names.csv --begin-date 2020-01-01 --end-date 2021-06-15 --token YOUR_TOKEN  > domain_name_summary.csv
    """,
    variadic_input={"dummy_column": "domain_name", "item_label": "domain name"},
    arguments=[
        *DATE_ARGUMENTS,
    ],
)

BUZZSUMO_COMMAND = command(
    "buzzsumo",
    "minet.cli.buzzsumo",
    title="Minet Buzzsumo Command",
    description="""
        Gather data from the BuzzSumo APIs easily and efficiently.
    """,
    aliases=["bz"],
    common_arguments=[
        {
            "flags": ["-t", "--token"],
            "help": "BuzzSumo API token.",
            "action": ConfigAction,
            "rc_key": ["buzzsumo", "token"],
            "required": True,
        }
    ],
    subcommands=[
        BUZZSUMO_LIMIT_SUBCOMMAND,
        BUZZSUMO_DOMAIN_COMMAND,
        BUZZSUMO_DOMAIN_SUMMARY_COMMAND,
    ],
)



# =============================================================================
# Minet Tiktok Constants
# =============================================================================
#
# General constants used throughout the Tiktok functions.
#
from urllib3 import Timeout

TIKTOK_URL = "https://www.tiktok.com"
TIKTOK_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
TIKTOK_DEFAULT_THROTTLE = 3.0
TIKTOK_MIN_TIME_RETRYER = 5.0
TIKTOK_MAX_RANDOM_ADDENDUM = 1.0

TIKTOK_VIDEO_CSV_HEADERS = [
    "id",
    "description",
    "timestamp",
    "utc_time",
    "original_item",
    "video_id",
    "video_height",
    "video_width",
    "video_duration",
    "video_ratio",
    "video_cover",
    "video_origin_cover",
    "video_dynamic_cover",
    # Access denied
    # "video_download_addr",
    "video_reflow_cover",
    "video_bitrate",
    "video_format",
    "video_quality",
    "author_id",
    "author_unique_id",
    "author_nickname",
    "author_avatar",
    "author_signature",
    "author_verified",
    "author_secret",
    "author_following_count",
    "author_follower_count",
    "author_heart_count",
    "author_video_count",
    "author_digg_count",
    "music_id",
    "music_title",
    "music_author_name",
    "music_original",
    "music_play_url",
    "music_duration",
    "music_album",
    "digg_count",
    "share_count",
    "comment_count",
    "play_count",
    "duet_from_video_id",
    "duet_from_user_id",
    "duet_from_user_name",
    "duet_from_user_is_commerce",
    "challenges_titles",
    "challenges_descriptions",
    "stickers_texts",
    "effect_stickers_names",
    "duet_enabled",
    "stitch_enabled",
    "share_enabled",
    "is_ad",
]



# =============================================================================
# Minet Facebook Post CLI Action
# =============================================================================
#
# Logic of the `fb post` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.facebook.utils import (
    with_facebook_fatal_errors,
    print_translation_warning_if_needed,
)
from minet.facebook import FacebookMobileScraper
from minet.facebook.constants import (
    FACEBOOK_POST_WITH_REACTIONS_CSV_HEADERS,
)
from minet.facebook.exceptions import (
    FacebookInvalidTargetError,
    FacebookNotPostError,
    FacebookWatchError,
)

CATCH = {
    FacebookInvalidTargetError: "Url is not available (link broken, private content, content removed, or scraping error): [info]{item}[/info]",
    FacebookNotPostError: "Url is not a post (may be an incorrect url): [info]{item}[/info]",
    FacebookWatchError: "Url leads to Facebook Watch and cannot be processed: [info]{item}[/info]",
}


@with_facebook_fatal_errors
@with_enricher_and_loading_bar(
    headers=FACEBOOK_POST_WITH_REACTIONS_CSV_HEADERS,
    title="Scraping posts",
    unit="posts",
)
def action(cli_args, enricher, loading_bar):
    scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)

    translated_langs = set()

    for i, row, url in enricher.enumerate_cells(cli_args.column, with_rows=True):
        with loading_bar.step(url, index=i, catch=CATCH):
            post = scraper.post(url)
            print_translation_warning_if_needed(loading_bar, translated_langs, post)
            enricher.writerow(row, post.as_csv_row())



# =============================================================================
# Minet CrowdTangle Posts
# =============================================================================
#
# Function related to posts fetching.
#
from minet.crowdtangle.utils import make_paginated_iterator
from minet.crowdtangle.formatters import format_post

URL_TEMPLATE = (
    "https://api.crowdtangle.com/posts?count=100&sortBy=%(sort_by)s&token=%(token)s"
)


def url_forge(
    token=None,
    sort_by=None,
    language=None,
    start_date=None,
    end_date=None,
    list_ids=None,
    **kwargs
):
    base_url = URL_TEMPLATE % {"sort_by": sort_by, "token": token}

    if language:
        base_url += "&language=%s" % language

    if start_date:
        base_url += "&startDate=%s" % start_date

    if end_date:
        base_url += "&endDate=%s" % end_date

    if list_ids:
        base_url += "&listIds=%s" % ",".join(list_ids)

    return base_url


crowdtangle_posts = make_paginated_iterator(
    url_forge, item_key="posts", formatter=format_post
)



# =============================================================================
# Minet CrowdTangle CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle actions.
#
import casanova
import casanova.ndjson as ndjson

from minet.cli.utils import print_err, with_fatal_errors
from minet.cli.loading_bar import LoadingBar
from minet.cli.exceptions import InvalidArgumentsError
from minet.crowdtangle import CrowdTangleAPIClient
from minet.crowdtangle.exceptions import CrowdTangleInvalidTokenError

FATAL_ERRORS = {
    CrowdTangleInvalidTokenError: [
        "Your API token is invalid.",
        "Check that you indicated a valid one using the `-t/--token` argument.",
    ]
}


def with_crowdtangle_fatal_errors(fn):
    return with_fatal_errors(FATAL_ERRORS)(fn)


def with_crowdtangle_client(fn):
    def wrapped(cli_args, *args, **kwargs):
        return fn(
            cli_args,
            *args,
            **{
                "client": CrowdTangleAPIClient(
                    cli_args.token, rate_limit=cli_args.rate_limit
                ),
                **kwargs,
            }
        )

    return wrapped


def with_crowdtangle_utilities(fn):
    return with_crowdtangle_fatal_errors(with_crowdtangle_client(fn))


def make_paginated_action(
    method_name, item_name, csv_headers, get_args=None, announce=None
):
    @with_crowdtangle_fatal_errors
    def action(cli_args):
        resume = getattr(cli_args, "resume", False)

        # Validation
        if resume:
            if cli_args.sort_by != "date":
                raise InvalidArgumentsError(
                    "Cannot --resume if --sort_by is not `date`."
                )

            if cli_args.format != "csv":
                raise InvalidArgumentsError("Cannot --resume jsonl format yet.")

        if cli_args.format == "csv":
            fieldnames = csv_headers(cli_args) if callable(csv_headers) else csv_headers
            writer = casanova.writer(cli_args.output, fieldnames=fieldnames)
        else:
            writer = ndjson.writer(cli_args.output)

        # Acquiring state from resumer
        if getattr(cli_args, "resume", False):
            last_date = cli_args.output.pop_state()

            if last_date is not None:
                cli_args.end_date = last_date.replace(" ", "T")
                print_err("Resuming from: %s" % cli_args.end_date)

        if callable(announce):
            print_err(announce(cli_args))

        # Loading bar
        loading_bar = LoadingBar(
            title="Fetching %s" % item_name, unit=item_name, total=cli_args.limit
        )

        args = []

        if callable(get_args):
            args = get_args(cli_args)

        client = CrowdTangleAPIClient(cli_args.token, rate_limit=cli_args.rate_limit)

        create_iterator = getattr(client, method_name)

        iterator = create_iterator(
            *args,
            limit=cli_args.limit,
            raw=cli_args.format != "csv",
            per_call=True,
            detailed=True,
            namespace=cli_args
        )

        with loading_bar:
            for _, items in iterator:
                loading_bar.advance(len(items))

                # if details is not None:
                #     loading_bar.update_stats(**details)

                for item in items:
                    if cli_args.format == "csv":
                        item = item.as_csv_row()

                    writer.writerow(item)

    return action



from minet.crawl.focus import FocusCrawlInfo, FocusSpider
from minet.crawl.types import (
    CrawlResult,
    AnyCrawlResult,
    SuccessfulCrawlResult,
    ErroredCrawlResult,
    CrawlJob,
    CrawlTarget,
    UrlOrCrawlTarget,
)
from minet.crawl.state import CrawlerState
from minet.crawl.spiders import Spider, SpiderResult, SpiderNextTargets, FunctionSpider
from minet.crawl.crawler import Crawler, AnySpider, SpiderDeclaration



from casanova import RowCountResumer

from minet.cli.exceptions import InvalidArgumentsError
from minet.cli.argparse import command


def resolve_pageviews_args(cli_args):
    if cli_args.lang is None and cli_args.lang_column is None:
        raise InvalidArgumentsError(
            "This command requires --lang or --lang-column to be given."
        )

    if cli_args.resume and not cli_args.sum:
        raise InvalidArgumentsError("Can only --resume with --sum.")


WIKIPEDIA_PAGEVIEWS_SUBCOMMAND = command(
    "pageviews",
    "minet.cli.wikipedia.pageviews",
    title="Minet Wikipedia Pageviews Command",
    description="""
        Command using the Wikimedia REST API to collect
        pageviews for a given amount of Wikipedia pages.

        See API documentation here for more details:
        https://wikitech.wikimedia.org/wiki/Analytics/AQS/Pageviews
    """,
    variadic_input={"dummy_column": "page"},
    resumer=RowCountResumer,
    resumer_epilog="Will only work with --sum.",
    arguments=[
        {
            "flag": "--start-date",
            "help": "Starting date. Must be of format YYYYMMDD (e.g. 20151031) or YYYYMMDDHH (e.g. 2015103100)",
            "required": True,
        },
        {
            "flag": "--end-date",
            "help": "End date. Must be of format YYYYMMDD (e.g. 20151031) or YYYYMMDDHH (e.g. 2015103100)",
            "required": True,
        },
        {
            "flag": "--agent",
            "help": "Get pageviews by target agent.",
            "default": "all-agents",
        },
        {
            "flag": "--access",
            "help": "Get pageviews by access.",
            "default": "all-access",
        },
        {
            "flags": ["-t", "--threads"],
            "help": "Number of threads to use.",
            "type": int,
            "default": 10,
        },
        {
            "flag": "--granularity",
            "help": "Pageviews granularity.",
            "default": "monthly",
        },
        {
            "flag": "--sum",
            "help": "Whether to sum the collected pageviews rather than outputting them by timestamp.",
            "action": "store_true",
        },
        {"flag": "--lang", "help": "Lang for the given pages."},
        {"flag": "--lang-column", "help": "Name of a CSV column containing page lang."},
    ],
)

WIKIPEDIA_COMMAND = command(
    "wikipedia",
    "minet.cli.wikipedia",
    title="Minet Wikipedia Command",
    aliases=["wiki"],
    subcommands=[WIKIPEDIA_PAGEVIEWS_SUBCOMMAND],
)



import sys
from typing import Union
from os import PathLike
from io import FileIO
from bs4 import BeautifulSoup

# NOTE: yes this is repetitive, but mypy only understands this, don't ask questions...
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if sys.version_info >= (3, 10):
    from typing import TypeGuard, Concatenate, ParamSpec
else:
    from typing_extensions import TypeGuard, Concatenate, ParamSpec

if sys.version_info >= (3, 11):
    from typing import TypedDict, Required, NotRequired
else:
    from typing_extensions import TypedDict, Required, NotRequired

AnyPath = Union[str, PathLike]
AnyFileTarget = Union[AnyPath, FileIO]
AnyScrapableTarget = Union[str, BeautifulSoup]



# =============================================================================
# Minet Custom Exceptions
# =============================================================================
#
# Collection of handy custom exceptions.
#
from minet.utils import message_flatmap


# Base minet error
class MinetError(Exception):
    def __init__(self, message=None):
        if message:
            message = message_flatmap(message)

        self.message = message or ""

        if not message:
            super().__init__()
        else:
            super().__init__(message)

    def __repr__(self):
        representation = "<" + self.__class__.__name__

        for k in dir(self):
            if k.startswith("_") or k == "args" or "traceback" in k:
                continue

            v = getattr(self, k)

            if v is None:
                continue

            if isinstance(v, BaseException):
                representation += " {k}=<{name}>".format(k=k, name=v.__class__.__name__)
                continue

            if isinstance(v, str) and len(v) > 30:
                v = v[:29] + "…"

            representation += " {k}={v!r}".format(k=k, v=v)

        representation += ">"

        return representation


# General errors
class UnknownEncodingError(MinetError):
    pass


class CouldNotInferEncodingError(MinetError):
    pass


# Cookie errors
class UnknownBrowserError(MinetError):
    def __init__(self, browser: str):
        super().__init__("Unknown browser: %s" % browser)
        self.browser = browser


class CookieGrabbingError(MinetError):
    def __init__(self, browser: str, reason: Exception):
        super().__init__("Could not grab cookies from: %s" % browser)
        self.browser = browser
        self.reason = reason


# Miscellaneous HTTP errors
class InvalidURLError(MinetError):
    def __init__(self, url: str):
        self.url = url
        super().__init__(url)


class InvalidStatusError(MinetError):
    def __init__(self, status: int):
        super().__init__(str(status))
        self.status = status


class HTTPCallbackError(MinetError):
    def __init__(self, reason: Exception):
        super().__init__("HTTPCallbackError: " + str(reason))
        self.reason = reason

    def unwrap(self, allow) -> Exception:
        if not isinstance(self.reason, allow):
            raise self.reason

        return self.reason


class CancelledRequestError(MinetError):
    pass


class FinalTimeoutError(MinetError):
    pass


# Redirection errors
class RedirectError(MinetError):
    pass


class MaxRedirectsError(RedirectError):
    pass


class InfiniteRedirectsError(RedirectError):
    pass


class SelfRedirectError(InfiniteRedirectsError):
    pass


class InvalidRedirectError(RedirectError):
    pass


class BadlyEncodedLocationHeaderError(RedirectError):
    pass


# Definition errors
class DefinitionInvalidFormatError(MinetError):
    pass


# Crawling errors
class CrawlError(MinetError):
    pass


class UnknownSpiderError(CrawlError):
    def __init__(self, spider: str):
        super().__init__(spider)
        self.spider = spider


# Extraction errors
class TrafilaturaError(MinetError):
    def __init__(self, message=None, reason=None):
        super().__init__(message)
        self.reason = reason


# Filesystem errors
class FilenameFormattingError(MinetError):
    def __init__(self, message=None, reason=None, template=None):
        super().__init__(message)
        self.reason = reason
        self.template = template



# =============================================================================
# Minet Facebook Constants
# =============================================================================
#
# General constants used throughout the Facebook namespace.
#
from collections import OrderedDict

from minet.rate_limiting import RateLimiterState

FACEBOOK_URL = "https://www.facebook.com"
FACEBOOK_MOBILE_URL = "https://m.facebook.com"

FACEBOOK_MOBILE_DEFAULT_THROTTLE = 2.0
FACEBOOK_WEB_DEFAULT_THROTTLE = 20.0

FACEBOOK_MOBILE_RATE_LIMITER_STATE = RateLimiterState(
    1, FACEBOOK_MOBILE_DEFAULT_THROTTLE
)
FACEBOOK_WEB_RATE_LIMITER_STATE = RateLimiterState(1, FACEBOOK_WEB_DEFAULT_THROTTLE)

FACEBOOK_USER_CSV_HEADERS = ["user_label", "user_id", "user_handle", "user_url"]

FACEBOOK_COMMENT_CSV_HEADERS = [
    "post_id",
    "id",
    "user_id",
    "user_handle",
    "user_url",
    "user_label",
    "text",
    "html",
    "formatted_date",
    "date",
    "reactions",
    "replies",
    "in_reply_to",
]

FACEBOOK_POST_CSV_HEADERS = [
    "url",
    "user_id",
    "user_handle",
    "user_url",
    "user_label",
    "text",
    "html",
    "translated_text",
    "translated_html",
    "translated_from",
    "formatted_date",
    "date",
    "reactions",
    "comments",
]

FACEBOOK_POST_STATS_CSV_HEADERS = [
    "error",
    "canonical",
    "account_name",
    "timestamp",
    "time",
    "link",
    "aria_label",
    "text",
    "share_count",
    "comment_count",
    "reaction_count",
    "video_view_count",
]

FACEBOOK_REACTION_KEYS = OrderedDict(
    {
        1: "like",
        2: "love",
        3: "wow",
        4: "haha",
        7: "sad",
        8: "angry",
        11: "thankful",
        12: "pride",
        16: "care",
    }
)

FACEBOOK_POST_WITH_REACTIONS_CSV_HEADERS = FACEBOOK_POST_CSV_HEADERS + [
    "reactions_types"
]



from minet.cli.cookies import COOKIES_COMMAND
from minet.cli.crawl import CRAWL_COMMAND, FOCUS_CRAWL_COMMAND
from minet.cli.dump import DUMP_QUEUE_COMMAND, DUMP_URLS_COMMAND, DUMP_GRAPH_COMMAND
from minet.cli.extract import EXTRACT_COMMAND
from minet.cli.fetch import FETCH_COMMAND, RESOLVE_COMMAND
from minet.cli.scrape import SCRAPE_COMMAND
from minet.cli.url_extract import URL_EXTRACT_COMMAND
from minet.cli.url_join import URL_JOIN_COMMAND
from minet.cli.url_parse import URL_PARSE_COMMAND

from minet.cli.buzzsumo import BUZZSUMO_COMMAND
from minet.cli.crowdtangle import CROWDTANGLE_COMMAND
from minet.cli.facebook import FACEBOOK_COMMAND
from minet.cli.google import GOOGLE_COMMAND
from minet.cli.hyphe import HYPHE_COMMAND
from minet.cli.instagram import INSTAGRAM_COMMAND
from minet.cli.mediacloud import MEDIACLOUD_COMMAND
from minet.cli.telegram import TELEGRAM_COMMAND
from minet.cli.tiktok import TIKTOK_COMMAND
from minet.cli.twitter import TWITTER_COMMAND
from minet.cli.wikipedia import WIKIPEDIA_COMMAND
from minet.cli.youtube import YOUTUBE_COMMAND

MINET_COMMANDS = [
    COOKIES_COMMAND,
    CRAWL_COMMAND,
    FOCUS_CRAWL_COMMAND,
    DUMP_QUEUE_COMMAND,
    DUMP_URLS_COMMAND,
    DUMP_GRAPH_COMMAND,
    EXTRACT_COMMAND,
    FETCH_COMMAND,
    RESOLVE_COMMAND,
    SCRAPE_COMMAND,
    URL_EXTRACT_COMMAND,
    URL_JOIN_COMMAND,
    URL_PARSE_COMMAND,
    #
    BUZZSUMO_COMMAND,
    CROWDTANGLE_COMMAND,
    FACEBOOK_COMMAND,
    GOOGLE_COMMAND,
    HYPHE_COMMAND,
    INSTAGRAM_COMMAND,
    MEDIACLOUD_COMMAND,
    TELEGRAM_COMMAND,
    TIKTOK_COMMAND,
    TWITTER_COMMAND,
    WIKIPEDIA_COMMAND,
    YOUTUBE_COMMAND,
]



# =============================================================================
# Minet YouTube Constants
# =============================================================================
#
# General constants used throughout the YouTube functions.
#

YOUTUBE_API_BASE_URL = "https://www.googleapis.com/youtube/v3"
YOUTUBE_API_MAX_VIDEOS_PER_CALL = 50
YOUTUBE_API_MAX_CHANNELS_PER_CALL = 50
YOUTUBE_API_MAX_COMMENTS_PER_CALL = 100

YOUTUBE_API_SEARCH_ORDERS = {
    "relevance",
    "date",
    "rating",
    "viewCount",
    "title",
    "videoCount",
}

YOUTUBE_API_DEFAULT_SEARCH_ORDER = "relevance"

YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS = [
    "video_id",
    "published_at",
    "channel_id",
    "title",
    "description",
    "channel_title",
]

YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS = YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS + [
    "position"
]

YOUTUBE_VIDEO_CSV_HEADERS = YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS + [
    "view_count",
    "like_count",
    # 'dislike_count',
    # This property is deprecated since december 13th 2021.
    # 'favorite_count',
    # This property has been deprecated by YouTube in 2015. The property's value is now always set to 0.
    "comment_count",
    "duration",
    "has_caption",
]

YOUTUBE_COMMENT_CSV_HEADERS = [
    "video_id",
    "comment_id",
    "author_name",
    "author_channel_id",
    "text",
    "like_count",
    "published_at",
    "updated_at",
    "reply_count",
    "parent_comment_id",
]

YOUTUBE_CAPTIONS_CSV_HEADERS = ["lang", "generated", "start", "duration", "text"]

YOUTUBE_CHANNEL_CSV_HEADERS = [
    "id",
    "title",
    "description",
    "custom_url",
    "published_at",
    "thumbnail",
    "default_language",
    "country",
    "id_playlists_videos",
    "view_count",
    "hidden_subscriber_count",
    "subscriber_count",
    "video_count",
    "topic_ids",
    "topic_categories",
    "topic_keywords",
    "privacy_status",
    "made_for_kids",
    "long_uploads_status",
    "keywords",
    "moderate_comments",
    "unsubscribed_trailer",
    "banner_external_url",
]



from minet.cli.utils import with_fatal_errors
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidTokenError,
    BuzzSumoInvalidQueryError,
)

FATAL_ERRORS = {BuzzSumoInvalidTokenError: "Your BuzzSumo token is invalid!"}


def fatal_errors_hook(_, e):
    if isinstance(e, BuzzSumoInvalidQueryError):
        return ["Invalid query: %s" % e.url, "Message from the API: %s" % e]

    return FATAL_ERRORS.get(type(e))


def with_buzzsumo_fatal_errors(fn):
    return with_fatal_errors(fatal_errors_hook)(fn)



# =============================================================================
# Minet Scraper Straining
# =============================================================================
#
# Function related to automatic bs4 SoupStrainer creation from simple
# CSS selectors.
#
import re
from bs4 import SoupStrainer
from soupsieve import SelectorSyntaxError
from soupsieve.css_parser import CSSParser

from minet.scrape.exceptions import CSSSelectorTooComplex, InvalidCSSSelectorError

WHITESPACE_RE = re.compile(r"\s+")


def match_selector(selector, tag, attrs):
    # Checking tag name
    if selector.tag.name != "*" and tag != selector.tag.name:
        return False

    # Checking id
    if selector.ids and attrs.get("id") not in selector.ids:
        return False

    # Checking class
    if selector.classes:
        classes = WHITESPACE_RE.split(attrs.get("class", ""))

        # NOTE: this is quadratic but probably faster than using sets in most cases?
        if not any(c in selector.classes for c in classes):
            return False

    # Checking attributes
    for target in selector.attributes:
        if target.attribute not in attrs:
            return False

        if target.pattern:
            if not target.pattern.match(attrs[target.attribute]):
                return False

    return True


def strainer_from_css(css: str, ignore_relations: bool = False) -> SoupStrainer:
    try:
        selector_list = CSSParser(css).process_selectors()
    except (SelectorSyntaxError, NotImplementedError) as e:
        raise InvalidCSSSelectorError(reason=e)

    usable_selectors = []

    for selector in selector_list:
        if selector.selectors or selector.nth:
            raise CSSSelectorTooComplex

        if not ignore_relations:
            if len(selector.relation) != 0:
                raise CSSSelectorTooComplex

            usable_selectors.append(selector)
        else:
            while selector.relation:
                selector = selector.relation[0]

            usable_selectors.append(selector)

    def strainer_function(tag, attrs):
        return any(
            match_selector(selector, tag, attrs) for selector in usable_selectors
        )

    strainer = SoupStrainer(strainer_function)

    setattr(strainer, "css", css)

    return strainer



# =============================================================================
# Minet Facebook Posts CLI Action
# =============================================================================
#
# Logic of the `fb posts` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.facebook.utils import (
    with_facebook_fatal_errors,
    print_translation_warning_if_needed,
)
from minet.facebook import FacebookMobileScraper
from minet.facebook.constants import FACEBOOK_POST_CSV_HEADERS
from minet.facebook.exceptions import FacebookInvalidTargetError


@with_facebook_fatal_errors
@with_enricher_and_loading_bar(
    headers=FACEBOOK_POST_CSV_HEADERS,
    title="Scraping group posts",
    unit="groups",
    nested=True,
    sub_unit="posts",
)
def action(cli_args, enricher, loading_bar):
    scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)

    translated_langs = set()

    for i, row, url in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step(url):
            try:
                posts = scraper.posts(url)
            except FacebookInvalidTargetError:
                loading_bar.print(
                    "Given url (line %i) is probably not a Facebook group: %s"
                    % (i, url)
                )
                continue

            for post in posts:
                print_translation_warning_if_needed(loading_bar, translated_langs, post)
                loading_bar.nested_advance()
                enricher.writerow(row, post.as_csv_row())



from typing import Optional, List, Any
from minet.types import AnyScrapableTarget

from bs4 import SoupStrainer, BeautifulSoup
from casanova import CSVSerializer
from urllib.parse import urljoin
from ural import should_follow_href

from minet.scrape.analysis import ScraperAnalysisOutputType
from minet.scrape.utils import ensure_soup
from minet.scrape.mixin import ScraperMixin


class NamedScraper(ScraperMixin):
    name: str
    fieldnames: List[str]
    plural: bool
    output_type: ScraperAnalysisOutputType
    strainer: Optional[SoupStrainer]
    serializer = CSVSerializer()

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        raise NotImplementedError

    def __call__(self, html: AnyScrapableTarget, context=None) -> Any:
        soup = ensure_soup(html, strainer=self.strainer)
        return self.scrape(soup, context=context)


class TitleScraper(NamedScraper):
    name = "title"
    fieldnames = ["page_title"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="title")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        title_elem = soup.find(name="title")

        if title_elem is None:
            return None

        return title_elem.get_text().strip()


class CanonicalScraper(NamedScraper):
    name = "canonical"
    fieldnames = ["canonical_url"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="link", attrs={"rel": "canonical"})

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        link_elem = soup.select_one("link[rel=canonical][href]")

        if link_elem is None:
            return None

        url = link_elem.get("href")

        if url is None:
            return None

        url = url.strip()

        if not url:
            return None

        return url


class UrlsScraper(NamedScraper):
    name = "urls"
    fieldnames = ["scraped_url"]
    plural = True
    output_type = "list"
    strainer = SoupStrainer(name="a")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        a_elems = soup.select("a[href]")
        base_url = context.get("url") if context is not None else None

        urls = []

        for a in a_elems:
            url = a.get("href")

            if url is None:
                continue

            url = url.strip()

            if not url:
                continue

            if not should_follow_href(url):
                continue

            if base_url:
                url = urljoin(base_url, url)

            urls.append(url)

        return urls


TYPICAL_SCRAPERS = {s.name: s for s in [TitleScraper, CanonicalScraper, UrlsScraper]}
